==> ksqldb-engine-7.1.9-sources.jar: io/confluent/ksql/function/udaf/offset/EarliestByOffset.java <==
/*
 * Copyright 2020 Confluent Inc.
 *
 * Licensed under the Confluent Community License (the "License"; you may not use
 * this file except in compliance with the License. You may obtain a copy of the
 * License at
 *
 * http://www.confluent.io/confluent-community-license
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
 * WARRANTIES OF ANY KIND, either express or implied.  See the License for the
 * specific language governing permissions and limitations under the License.
 */

package io.confluent.ksql.function.udaf.offset;

import static io.confluent.ksql.function.udaf.offset.KudafByOffsetUtils.INTERMEDIATE_STRUCT_COMPARATOR;
import static io.confluent.ksql.function.udaf.offset.KudafByOffsetUtils.STRUCT_BOOLEAN;
import static io.confluent.ksql.function.udaf.offset.KudafByOffsetUtils.STRUCT_BYTES;
import static io.confluent.ksql.function.udaf.offset.KudafByOffsetUtils.STRUCT_DATE;
import static io.confluent.ksql.function.udaf.offset.KudafByOffsetUtils.STRUCT_DOUBLE;
import static io.confluent.ksql.function.udaf.offset.KudafByOffsetUtils.STRUCT_INTEGER;
import static io.confluent.ksql.function.udaf.offset.KudafByOffsetUtils.STRUCT_LONG;
import static io.confluent.ksql.function.udaf.offset.KudafByOffsetUtils.STRUCT_STRING;
import static io.confluent.ksql.function.udaf.offset.KudafByOffsetUtils.STRUCT_TIME;
import static io.confluent.ksql.function.udaf.offset.KudafByOffsetUtils.STRUCT_TIMESTAMP;
import static io.confluent.ksql.function.udaf.offset.KudafByOffsetUtils.VAL_FIELD;

import com.google.common.annotations.VisibleForTesting;
import io.confluent.ksql.function.KsqlFunctionException;
import io.confluent.ksql.function.udaf.Udaf;
import io.confluent.ksql.function.udaf.UdafDescription;
import io.confluent.ksql.function.udaf.UdafFactory;
import io.confluent.ksql.util.KsqlConstants;
import java.nio.ByteBuffer;
import java.sql.Date;
import java.sql.Time;
import java.sql.Timestamp;
import java.util.ArrayList;
import java.util.List;
import java.util.concurrent.atomic.AtomicLong;
import java.util.stream.Collectors;
import org.apache.kafka.connect.data.Schema;
import org.apache.kafka.connect.data.Struct;

@UdafDescription(
    name = "EARLIEST_BY_OFFSET",
    description = EarliestByOffset.DESCRIPTION,
    author = KsqlConstants.CONFLUENT_AUTHOR
)
public final class EarliestByOffset {

  static final String DESCRIPTION =
      "This function returns the oldest N values for the column, computed by offset.";

  private EarliestByOffset() {
  }

  static AtomicLong sequence = new AtomicLong();

  @UdafFactory(description = "return the earliest value of an integer column",
      aggregateSchema = "STRUCT<SEQ BIGINT, VAL INT>")
  public static Udaf<Integer, Struct, Integer> earliestInteger() {
    return earliestInteger(true);
  }

  @UdafFactory(description = "return the earliest value of an integer column",
      aggregateSchema = "STRUCT<SEQ BIGINT, VAL INT>")
  public static Udaf<Integer, Struct, Integer> earliestInteger(final boolean ignoreNulls) {
    return earliest(STRUCT_INTEGER, ignoreNulls);
  }

  @UdafFactory(description = "return the earliest N values of an integer column",
      aggregateSchema = "ARRAY<STRUCT<SEQ BIGINT, VAL INT>>")
  public static Udaf<Integer, List<Struct>, List<Integer>> earliestIntegers(final int earliestN) {
    return earliestIntegers(earliestN, true);
  }

  @UdafFactory(description = "return the earliest N values of an integer column",
      aggregateSchema = "ARRAY<STRUCT<SEQ BIGINT, VAL INT>>")
  public static Udaf<Integer, List<Struct>, List<Integer>> earliestIntegers(
      final int earliestN,
      final boolean ignoreNulls
  ) {
    return earliestN(STRUCT_INTEGER, earliestN, ignoreNulls);
  }

  @UdafFactory(description = "return the earliest value of an big integer column",
      aggregateSchema = "STRUCT<SEQ BIGINT, VAL BIGINT>")
  public static Udaf<Long, Struct, Long> earliestLong() {
    return earliestLong(true);
  }

  @UdafFactory(description = "return the earliest value of an big integer column",
      aggregateSchema = "STRUCT<SEQ BIGINT, VAL BIGINT>")
  public static Udaf<Long, Struct, Long> earliestLong(final boolean ignoreNulls) {
    return earliest(STRUCT_LONG, ignoreNulls);
  }

  @UdafFactory(description = "return the earliest N values of an long column",
      aggregateSchema = "ARRAY<STRUCT<SEQ BIGINT, VAL BIGINT>>")
  public static Udaf<Long, List<Struct>, List<Long>> earliestLongs(final int earliestN) {
    return earliestLongs(earliestN, true);
  }

  @UdafFactory(description = "return the earliest N values of an long column",
      aggregateSchema = "ARRAY<STRUCT<SEQ BIGINT, VAL BIGINT>>")
  public static Udaf<Long, List<Struct>, List<Long>> earliestLongs(
      final int earliestN,
      final boolean ignoreNulls
  ) {
    return earliestN(STRUCT_LONG, earliestN, ignoreNulls);
  }

  @UdafFactory(description = "return the earliest value of a double column",
      aggregateSchema = "STRUCT<SEQ BIGINT, VAL DOUBLE>")
  public static Udaf<Double, Struct, Double> earliestDouble() {
    return earliestDouble(true);
  }

  @UdafFactory(description = "return the earliest value of a double column",
      aggregateSchema = "STRUCT<SEQ BIGINT, VAL DOUBLE>")
  public static Udaf<Double, Struct, Double> earliestDouble(final boolean ignoreNulls) {
    return earliest(STRUCT_DOUBLE, ignoreNulls);
  }

  @UdafFactory(description = "return the earliest N values of a double column",
      aggregateSchema = "ARRAY<STRUCT<SEQ BIGINT, VAL DOUBLE>>")
  public static Udaf<Double, List<Struct>, List<Double>> earliestDoubles(final int earliestN) {
    return earliestDoubles(earliestN, true);
  }

  @UdafFactory(description = "return the earliest N values of a double column",
      aggregateSchema = "ARRAY<STRUCT<SEQ BIGINT, VAL DOUBLE>>")
  public static Udaf<Double, List<Struct>, List<Double>> earliestDoubles(
      final int earliestN,
      final boolean ignoreNulls
  ) {
    return earliestN(STRUCT_DOUBLE, earliestN, ignoreNulls);
  }

  @UdafFactory(description = "return the earliest value of a boolean column",
      aggregateSchema = "STRUCT<SEQ BIGINT, VAL BOOLEAN>")
  public static Udaf<Boolean, Struct, Boolean> earliestBoolean() {
    return earliestBoolean(true);
  }

  @UdafFactory(description = "return the earliest value of a boolean column",
      aggregateSchema = "STRUCT<SEQ BIGINT, VAL BOOLEAN>")
  public static Udaf<Boolean, Struct, Boolean> earliestBoolean(final boolean ignoreNulls) {
    return earliest(STRUCT_BOOLEAN, ignoreNulls);
  }

  @UdafFactory(description = "return the earliest N values of a boolean column",
      aggregateSchema = "ARRAY<STRUCT<SEQ BIGINT, VAL BOOLEAN>>")
  public static Udaf<Boolean, List<Struct>, List<Boolean>> earliestBooleans(final int earliestN) {
    return earliestBooleans(earliestN, true);
  }

  @UdafFactory(description = "return the earliest N values of a boolean column",
      aggregateSchema = "ARRAY<STRUCT<SEQ BIGINT, VAL BOOLEAN>>")
  public static Udaf<Boolean, List<Struct>, List<Boolean>> earliestBooleans(
      final int earliestN,
      final boolean ignoreNulls
  ) {
    return earliestN(STRUCT_BOOLEAN, earliestN, ignoreNulls);
  }

  @UdafFactory(description = "return the earliest value of a string column",
      aggregateSchema = "STRUCT<SEQ BIGINT, VAL STRING>")
  public static Udaf<String, Struct, String> earliestString() {
    return earliestString(true);
  }

  @UdafFactory(description = "return the earliest value of a string column",
      aggregateSchema = "STRUCT<SEQ BIGINT, VAL STRING>")
  public static Udaf<String, Struct, String> earliestString(final boolean ignoreNulls) {
    return earliest(STRUCT_STRING, ignoreNulls);
  }

  @UdafFactory(description = "return the earliest N values of a string column",
      aggregateSchema = "ARRAY<STRUCT<SEQ BIGINT, VAL STRING>>")
  public static Udaf<String, List<Struct>, List<String>> earliestStrings(final int earliestN) {
    return earliestStrings(earliestN, true);
  }

  @UdafFactory(description = "return the earliest N values of a string column",
      aggregateSchema = "ARRAY<STRUCT<SEQ BIGINT, VAL STRING>>")
  public static Udaf<String, List<Struct>, List<String>> earliestStrings(
      final int earliestN,
      final boolean ignoreNulls
  ) {
    return earliestN(STRUCT_STRING, earliestN, ignoreNulls);
  }

  @UdafFactory(description = "return the earliest value of a timestamp column",
      aggregateSchema = "STRUCT<SEQ BIGINT, VAL TIMESTAMP>")
  public static Udaf<Timestamp, Struct, Timestamp> earliestTimestamp() {
    return earliestTimestamp(true);
  }

  @UdafFactory(description = "return the earliest value of a timestamp column",
      aggregateSchema = "STRUCT<SEQ BIGINT, VAL TIMESTAMP>")
  public static Udaf<Timestamp, Struct, Timestamp> earliestTimestamp(final boolean ignoreNulls) {
    return earliest(STRUCT_TIMESTAMP, ignoreNulls);
  }

  @UdafFactory(description = "return the earliest N values of a timestamp column",
      aggregateSchema = "ARRAY<STRUCT<SEQ BIGINT, VAL TIMESTAMP>>")
  public static Udaf<Timestamp, List<Struct>, List<Timestamp>> earliestTimestamps(
      final int earliestN) {
    return earliestTimestamps(earliestN, true);
  }

  @UdafFactory(description = "return the earliest N values of a timestamp column",
      aggregateSchema = "ARRAY<STRUCT<SEQ BIGINT, VAL TIMESTAMP>>")
  public static Udaf<Timestamp, List<Struct>, List<Timestamp>> earliestTimestamps(
      final int earliestN,
      final boolean ignoreNulls
  ) {
    return earliestN(STRUCT_TIMESTAMP, earliestN, ignoreNulls);
  }

  @UdafFactory(description = "return the earliest value of a time column",
      aggregateSchema = "STRUCT<SEQ BIGINT, VAL TIME>")
  public static Udaf<Time, Struct, Time> earliestTime() {
    return earliestTime(true);
  }

  @UdafFactory(description = "return the earliest value of a time column",
      aggregateSchema = "STRUCT<SEQ BIGINT, VAL TIME>")
  public static Udaf<Time, Struct, Time> earliestTime(final boolean ignoreNulls) {
    return earliest(STRUCT_TIME, ignoreNulls);
  }

  @UdafFactory(description = "return the earliest N values of a time column",
      aggregateSchema = "ARRAY<STRUCT<SEQ BIGINT, VAL TIME>>")
  public static Udaf<Time, List<Struct>, List<Time>> earliestTimes(final int earliestN) {
    return earliestTimes(earliestN, true);
  }

  @UdafFactory(description = "return the earliest N values of a time column",
      aggregateSchema = "ARRAY<STRUCT<SEQ BIGINT, VAL TIME>>")
  public static Udaf<Time, List<Struct>, List<Time>> earliestTimes(
      final int earliestN,
      final boolean ignoreNulls
  ) {
    return earliestN(STRUCT_TIME, earliestN, ignoreNulls);
  }

  @UdafFactory(description = "return the earliest value of a date column",
      aggregateSchema = "STRUCT<SEQ BIGINT, VAL DATE>")
  public static Udaf<Date, Struct, Date> earliestDate() {
    return earliestDate(true);
  }

  @UdafFactory(description = "return the earliest value of a date column",
      aggregateSchema = "STRUCT<SEQ BIGINT, VAL DATE>")
  public static Udaf<Date, Struct, Date> earliestDate(final boolean ignoreNulls) {
    return earliest(STRUCT_DATE, ignoreNulls);
  }

  @UdafFactory(description = "return the earliest N values of a date column",
      aggregateSchema = "ARRAY<STRUCT<SEQ BIGINT, VAL DATE>>")
  public static Udaf<Date, List<Struct>, List<Date>> earliestDates(final int earliestN) {
    return earliestDates(earliestN, true);
  }

  @UdafFactory(description = "return the earliest N values of a date column",
      aggregateSchema = "ARRAY<STRUCT<SEQ BIGINT, VAL DATE>>")
  public static Udaf<Date, List<Struct>, List<Date>> earliestDates(
      final int earliestN,
      final boolean ignoreNulls
  ) {
    return earliestN(STRUCT_DATE, earliestN, ignoreNulls);
  }

  @UdafFactory(description = "return the earliest value of a bytes column",
      aggregateSchema = "STRUCT<SEQ BIGINT, VAL BYTES>")
  public static Udaf<ByteBuffer, Struct, ByteBuffer> earliestBytes() {
    return earliestBytes(true);
  }

  @UdafFactory(description = "return the earliest value of a bytes column",
      aggregateSchema = "STRUCT<SEQ BIGINT, VAL BYTES>")
  public static Udaf<ByteBuffer, Struct, ByteBuffer> earliestBytes(final boolean ignoreNulls) {
    return earliest(STRUCT_BYTES, ignoreNulls);
  }

  @UdafFactory(description = "return the earliest N values of a bytes column",
      aggregateSchema = "ARRAY<STRUCT<SEQ BIGINT, VAL BYTES>>")
  public static Udaf<ByteBuffer, List<Struct>, List<ByteBuffer>> earliestBytes(
      final int earliestN) {
    return earliestBytes(earliestN, true);
  }

  @UdafFactory(description = "return the earliest N values of a bytes column",
      aggregateSchema = "ARRAY<STRUCT<SEQ BIGINT, VAL BYTES>>")
  public static Udaf<ByteBuffer, List<Struct>, List<ByteBuffer>> earliestBytes(
      final int earliestN,
      final boolean ignoreNulls
  ) {
    return earliestN(STRUCT_BYTES, earliestN, ignoreNulls);
  }

  @VisibleForTesting
  static <T> Struct createStruct(final Schema schema, final T val) {
    return KudafByOffsetUtils.createStruct(schema, generateSequence(), val);
  }

  private static long generateSequence() {
    return sequence.getAndIncrement();
  }

  @VisibleForTesting
  static <T> Udaf<T, Struct, T> earliest(
      final Schema structSchema,
      final boolean ignoreNulls
  ) {
    return new Udaf<T, Struct, T>() {

      @Override
      public Struct initialize() {
        return null;
      }

      @Override
      public Struct aggregate(final T current, final Struct aggregate) {
        if (aggregate != null) {
          return aggregate;
        }

        if (current == null && ignoreNulls) {
          return null;
        }

        return createStruct(structSchema, current);
      }

      @Override
      public Struct merge(final Struct aggOne, final Struct aggTwo) {
        if (aggOne == null) {
          return aggTwo;
        }

        if (aggTwo == null) {
          return aggOne;
        }

        // When merging we need some way of evaluating the "earliest' one.
        // We do this by keeping track of the sequence of when it was originally processed
        if (INTERMEDIATE_STRUCT_COMPARATOR.compare(aggOne, aggTwo) < 0) {
          return aggOne;
        } else {
          return aggTwo;
        }
      }

      @Override
      @SuppressWarnings("unchecked")
      public T map(final Struct agg) {
        if (agg == null) {
          return null;
        }

        return (T) agg.get(VAL_FIELD);
      }
    };
  }

  @VisibleForTesting
  static <T> Udaf<T, List<Struct>, List<T>> earliestN(
      final Schema structSchema,
      final int earliestN,
      final boolean ignoreNulls
  ) {
    if (earliestN <= 0) {
      throw new KsqlFunctionException("earliestN must be 1 or greater");
    }

    return new Udaf<T, List<Struct>, List<T>>() {
      @Override
      public List<Struct> initialize() {
        return new ArrayList<>(earliestN);
      }

      @Override
      public List<Struct> aggregate(final T current, final List<Struct> aggregate) {
        if (current == null && ignoreNulls) {
          return aggregate;
        }

        if (aggregate.size() < earliestN) {
          aggregate.add(createStruct(structSchema, current));
        }
        return aggregate;
      }

      @Override
      public List<Struct> merge(final List<Struct> aggOne, final List<Struct> aggTwo) {
        final List<Struct> merged = new ArrayList<>(aggOne.size() + aggTwo.size());
        merged.addAll(aggOne);
        merged.addAll(aggTwo);
        merged.sort(INTERMEDIATE_STRUCT_COMPARATOR);
        return merged.subList(0, Math.min(earliestN, merged.size()));
      }

      @Override
      @SuppressWarnings("unchecked")
      public List<T> map(final List<Struct> agg) {
        return (List<T>) agg.stream().map(s -> s.get(VAL_FIELD)).collect(Collectors.toList());
      }
    };
  }
}

==> ksqldb-engine-7.2.7-sources.jar: io/confluent/ksql/function/udaf/offset/EarliestByOffset.java <==
/*
 * Copyright 2020 Confluent Inc.
 *
 * Licensed under the Confluent Community License (the "License"; you may not use
 * this file except in compliance with the License. You may obtain a copy of the
 * License at
 *
 * http://www.confluent.io/confluent-community-license
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
 * WARRANTIES OF ANY KIND, either express or implied.  See the License for the
 * specific language governing permissions and limitations under the License.
 */

package io.confluent.ksql.function.udaf.offset;

import static io.confluent.ksql.function.udaf.offset.KudafByOffsetUtils.INTERMEDIATE_STRUCT_COMPARATOR;
import static io.confluent.ksql.function.udaf.offset.KudafByOffsetUtils.VAL_FIELD;

import com.google.common.annotations.VisibleForTesting;
import io.confluent.ksql.function.KsqlFunctionException;
import io.confluent.ksql.function.udaf.Udaf;
import io.confluent.ksql.function.udaf.UdafDescription;
import io.confluent.ksql.function.udaf.UdafFactory;
import io.confluent.ksql.schema.ksql.SchemaConverters;
import io.confluent.ksql.schema.ksql.SqlArgument;
import io.confluent.ksql.schema.ksql.types.SqlArray;
import io.confluent.ksql.schema.ksql.types.SqlType;
import io.confluent.ksql.util.KsqlConstants;
import java.util.ArrayList;
import java.util.List;
import java.util.Optional;
import java.util.concurrent.atomic.AtomicLong;
import java.util.stream.Collectors;
import org.apache.kafka.connect.data.Schema;
import org.apache.kafka.connect.data.Struct;

@UdafDescription(
    name = "EARLIEST_BY_OFFSET",
    description = EarliestByOffset.DESCRIPTION,
    author = KsqlConstants.CONFLUENT_AUTHOR
)
public final class EarliestByOffset {

  static final String DESCRIPTION =
      "This function returns the oldest N values for the column, computed by offset.";

  private EarliestByOffset() {
  }

  static final AtomicLong sequence = new AtomicLong();

  @UdafFactory(description = "return the earliest value of a column")
  public static <T> Udaf<T, Struct, T> earliest() {
    return earliest(true);
  }

  @UdafFactory(description = "return the earliest value of a column")
  public static <T> Udaf<T, Struct, T> earliest(final boolean ignoreNulls) {
    return earliestT(ignoreNulls);
  }

  @UdafFactory(description = "return the earliest N values of a column")
  public static <T> Udaf<T, List<Struct>, List<T>> earliest(final int earliestN) {
    return earliest(earliestN, true);
  }

  @UdafFactory(description = "return the earliest N values of a column")
  public static <T> Udaf<T, List<Struct>, List<T>> earliest(final int earliestN,
      final boolean ignoreNulls) {
    return earliestTN(earliestN, ignoreNulls);
  }

  @VisibleForTesting
  static <T> Struct createStruct(final Schema schema, final T val) {
    return KudafByOffsetUtils.createStruct(schema, generateSequence(), val);
  }

  private static long generateSequence() {
    return sequence.getAndIncrement();
  }

  @VisibleForTesting
  static <T> Udaf<T, Struct, T> earliestT(
      final boolean ignoreNulls
  ) {
    return new Udaf<T, Struct, T>() {
      Schema structSchema;
      SqlType aggregateType;
      SqlType returnType;

      @Override
      public void initializeTypeArguments(final List<SqlArgument> argTypeList) {
        returnType = argTypeList.get(0).getSqlTypeOrThrow();
        final Schema connectType =
            SchemaConverters.sqlToConnectConverter().toConnectSchema(returnType);
        structSchema = KudafByOffsetUtils.buildSchema(connectType);
        aggregateType = SchemaConverters.connectToSqlConverter().toSqlType(structSchema);
      }

      @Override
      public Optional<SqlType> getAggregateSqlType() {
        return Optional.of(aggregateType);
      }

      @Override
      public Optional<SqlType> getReturnSqlType() {
        return Optional.of(returnType);
      }

      @Override
      public Struct initialize() {
        return null;
      }

      @Override
      public Struct aggregate(final T current, final Struct aggregate) {
        if (aggregate != null) {
          return aggregate;
        }

        if (current == null && ignoreNulls) {
          return null;
        }

        return createStruct(structSchema, current);
      }

      @Override
      public Struct merge(final Struct aggOne, final Struct aggTwo) {
        if (aggOne == null) {
          return aggTwo;
        }

        if (aggTwo == null) {
          return aggOne;
        }

        // When merging we need some way of evaluating the "earliest" one.
        // We do this by keeping track of the sequence of when it was originally processed
        if (INTERMEDIATE_STRUCT_COMPARATOR.compare(aggOne, aggTwo) < 0) {
          return aggOne;
        } else {
          return aggTwo;
        }
      }

      @Override
      @SuppressWarnings("unchecked")
      public T map(final Struct agg) {
        if (agg == null) {
          return null;
        }

        return (T) agg.get(VAL_FIELD);
      }
    };
  }

  @VisibleForTesting
  static <T> Udaf<T, List<Struct>, List<T>> earliestTN(
      final int earliestN,
      final boolean ignoreNulls
  ) {
    if (earliestN <= 0) {
      throw new KsqlFunctionException("earliestN must be 1 or greater");
    }

    return new Udaf<T, List<Struct>, List<T>>() {
      Schema structSchema;
      SqlType aggregateType;
      SqlType returnType;

      @Override
      public void initializeTypeArguments(final List<SqlArgument> argTypeList) {
        final SqlType inputType = argTypeList.get(0).getSqlTypeOrThrow();
        final Schema connectType =
            SchemaConverters.sqlToConnectConverter().toConnectSchema(inputType);
        structSchema = KudafByOffsetUtils.buildSchema(connectType);
        aggregateType =
            SqlArray.of(SchemaConverters.connectToSqlConverter().toSqlType(structSchema));
        returnType = SqlArray.of(inputType);
      }

      @Override
      public Optional<SqlType> getAggregateSqlType() {
        return Optional.of(aggregateType);
      }

      @Override
      public Optional<SqlType> getReturnSqlType() {
        return Optional.of(returnType);
      }

      @Override
      public List<Struct> initialize() {
        return new ArrayList<>(earliestN);
      }

      @Override
      public List<Struct> aggregate(final T current, final List<Struct> aggregate) {
        if (current == null && ignoreNulls) {
          return aggregate;
        }

        if (aggregate.size() < earliestN) {
          aggregate.add(createStruct(structSchema, current));
        }
        return aggregate;
      }

      @Override
      public List<Struct> merge(final List<Struct> aggOne, final List<Struct> aggTwo) {
        final List<Struct> merged = new ArrayList<>(aggOne.size() + aggTwo.size());
        merged.addAll(aggOne);
        merged.addAll(aggTwo);
        merged.sort(INTERMEDIATE_STRUCT_COMPARATOR);
        return merged.subList(0, Math.min(earliestN, merged.size()));
      }

      @Override
      @SuppressWarnings("unchecked")
      public List<T> map(final List<Struct> agg) {
        return (List<T>) agg.stream().map(s -> s.get(VAL_FIELD)).collect(Collectors.toList());
      }
    };
  }
}
